/*
 * 	consumer.c
 *
 * Copyright (c) 2004 Evgeniy Polyakov <johnpol@2ka.mipt.ru>
 * 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/types.h>
#include <linux/list.h>
#include <linux/slab.h>
#include <linux/spinlock.h>
#include <linux/workqueue.h>
#include <linux/err.h>
#include <linux/mm.h>
#include <linux/completion.h>

#undef DEBUG
#include "acrypto.h"
#include "crypto_user.h"
#include "crypto_def.h"

#undef dprintk
#define dprintk(f, a...) printk(f, ##a)
//#define dprintk(f, a...) do {} while(0)

#define KEY_LENGTH		16
static char ckey[KEY_LENGTH];
static int key_length = sizeof(ckey);

static void ctest_callback(struct crypto_session_initializer *ci,
			   struct crypto_data *data);

static struct crypto_session_initializer ci = {
	.operation 	= CRYPTO_OP_ENCRYPT,
	.type 		= CRYPTO_TYPE_AES_128,
	.mode 		= CRYPTO_MODE_ECB,
	.priority 	= 4,
	.callback 	= ctest_callback,
};

static struct crypto_data cdata;

#define CSESSION_MAX	1000
static struct crypto_session *s;
static int watermark;
static struct timer_list ctimer;
static char data_str[4096] = "test message qwerty";

static void ctest_callback(struct crypto_session_initializer *ci,
			   struct crypto_data *data)
{
	int i, off, size, ssize;
	struct scatterlist *sg;
	unsigned char *ptr;

	watermark--;

	printk("%s: session %llu [%llu]\n", __func__, ci->id, ci->dev_id);
	dprintk("src=%s, len=%d.\n",
		((char *)page_address(data->sg_src[0].page)) + data->sg_src[0].offset,
		data->sg_src[0].length);

	ssize = 32;

	sg = &data->sg_key[0];
	ptr = (unsigned char *)page_address(sg->page);
	off = sg->offset;
	size = sg->length;

	dprintk("key[%d]=", size);
	for (i = 0; i < size; ++i)
		dprintk("0x%02x, ", ptr[i + off]);
	dprintk("\n");

	sg = &data->sg_src[0];
	ptr = (unsigned char *)page_address(sg->page);
	off = sg->offset;
	size = sg->length;

	dprintk("src[%d]=", size);
	for (i = 0; i < ssize; ++i)
		dprintk("0x%02x, ", ptr[i + off]);
	dprintk("\n");

	sg = &data->sg_dst[0];
	ptr = (unsigned char *)page_address(sg->page);
	off = sg->offset;
	size = sg->length;

	dprintk("dst[%d]=", size);
	for (i = 0; i < ssize; ++i)
		dprintk("0x%02x, ", ptr[i + off]);
	dprintk("\n");
}

int ctimer_func(void *data, int size, int op)
{
	u8 *ptr;

	if (size > PAGE_SIZE)
		size = PAGE_SIZE;

	ptr = (u8 *)page_address(cdata.sg_src[0].page) + cdata.sg_src[0].offset;
	memcpy(ptr, data, size);
	cdata.sg_src[0].length = size;
	cdata.sg_dst[0].length = size;

	ci.operation = op;
	s = crypto_session_alloc(&ci, &cdata);
	if (s)
		watermark++;
	else
		dprintk("allocation failed.\n");

	return (s) ? 0 : -EINVAL;
}

static int alloc_sg(struct scatterlist *sg, void *data, int size)
{
	sg->offset = 0;
	sg->page = alloc_pages(GFP_KERNEL, get_order(size));
	if (!sg->page) {
		printk(KERN_ERR "Failed to allocate page.\n");
		return -ENOMEM;
	}

	memset(page_address(sg->page), 0, PAGE_SIZE);

	if (data)
		memcpy(page_address(sg->page), data, size);

	sg->length = size;

	return 0;
}

static void ctimerf(unsigned long data)
{
	int size, err;

	printk("%s started\n", __func__);

	size = 4096;

	err = ctimer_func(data_str, size, CRYPTO_OP_ENCRYPT);
	err = ctimer_func(data_str, size, CRYPTO_OP_DECRYPT);

	if (!err)
		mod_timer(&ctimer, jiffies + HZ);
	else
		mod_timer(&ctimer, jiffies + HZ);

	printk("%s finished.\n", __func__);
}

int consumer_init(void)
{
	int err;

	err = crypto_user_alloc_crypto_data(&cdata, sizeof(data_str), sizeof(data_str), key_length, 1);
	if (err)
		return err;

	err = alloc_sg(&cdata.sg_src[0], NULL, PAGE_SIZE);
	if (err)
		goto err_out_return;

	err = alloc_sg(&cdata.sg_dst[0], NULL, PAGE_SIZE);
	if (err)
		goto err_out_src;

	err = alloc_sg(&cdata.sg_key[0], ckey, key_length);
	if (err)
		goto err_out_dst;

	err = alloc_sg(&cdata.sg_iv[0], NULL, PAGE_SIZE);
	if (err)
		goto err_out_key;

	cdata.priv_size = 256;//sizeof(struct aes_ctx);

	init_timer(&ctimer);
	ctimer.function = &ctimerf;
	ctimer.expires = jiffies + HZ;
	ctimer.data = 0;

	add_timer(&ctimer);

	return 0;

	__free_pages(cdata.sg_iv[0].page, get_order(1));
err_out_key:
	__free_pages(cdata.sg_key[0].page, get_order(key_length));
err_out_dst:
	__free_pages(cdata.sg_dst[0].page, get_order(1));
err_out_src:
	__free_pages(cdata.sg_src[0].page, get_order(1));
err_out_return:
	crypto_user_free_crypto_data(&cdata);
	
	return -ENODEV;
}

void consumer_fini(void)
{
	del_timer_sync(&ctimer);
	while (watermark) {
		set_current_state(TASK_UNINTERRUPTIBLE);
		schedule_timeout(HZ);

		printk(KERN_INFO "Waiting for sessions to be freed: watermark=%d.\n",
			watermark);
	}

	if (!cdata.sg_key[0].length)
		printk("BUG: key length is 0 in %s.\n", __func__);

	__free_pages(cdata.sg_iv[0].page, get_order(1));
	__free_pages(cdata.sg_key[0].page, get_order(key_length));
	__free_pages(cdata.sg_dst[0].page, get_order(1));
	__free_pages(cdata.sg_src[0].page, get_order(1));
	
	crypto_user_free_crypto_data(&cdata);

	printk(KERN_INFO "Test crypto module consumer is unloaded.\n");
}

module_init(consumer_init);
module_exit(consumer_fini);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Evgeniy Polyakov <johnpol@2ka.mipt.ru>");
MODULE_DESCRIPTION("Test crypto module consumer.");
